/*
 * Copyright 2010 by Adam Mayer	 <adam@makerbot.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 */

#ifndef HEATER_H
#define HEATER_H

#include "TemperatureSensor.hh"
#include "HeatingElement.hh"
#include "Pin.hh"
#include "PID.hh"
#include "Types.hh"
#include "Timeout.hh"

//#define MAX_VALID_TEMP 280
//#define MAX_HBP_TEMP   130

#ifdef HEATER_CUSTOM_LIMIT
const uint16_t Heater_Max_Temp[2] = HEATER_CUSTOM_LIMIT;
#else
const uint16_t Heater_Max_Temp[2] = {280,130};
#endif

#  define MAX_VALID_TEMP Heater_Max_Temp[0]
#  define MAX_HBP_TEMP   Heater_Max_Temp[1]


#define DEFAULT_P 7.0
#define DEFAULT_I 0.325
#define DEFAULT_D 36.0

enum HeaterFailMode{
	HEATER_FAIL_NONE = 0,
	HEATER_FAIL_NOT_PLUGGED_IN = 0x02,
	HEATER_FAIL_SOFTWARE_CUTOFF = 0x04,
	HEATER_FAIL_NOT_HEATING = 0x08,
	HEATER_FAIL_DROPPING_TEMP = 0x10,
	HEATER_FAIL_BAD_READS = 0x20
};



/// A heater object uses a #TemperatureSensor to control a #HeatingElement.
/// It is primarily a #PID controller, but enters a bypass mode when the setpoint
/// temperature is much higher than the current temperature. The #manage_temperature()
/// function must be called periodically for it to be able to maintain accurate
/// temperature control.
/// \ingroup SoftwareLibraries
class Heater
{
  private:
    TemperatureSensor& sensor;          ///< Sensor used to measure input temperature
    HeatingElement& element;            ///< Heating element used to produce an output
    Timeout next_pid_timeout;           ///< Timeout timer for PID loop (should be slower
                                        ///< or the same speed as sensor timeout)
    volatile int16_t current_temperature;       ///< Last known temperature reading
    int16_t startTemp;		///< start temperature when new target is set.  used to assess heating up progress
    int16_t paused_set_temperature;		///< we record the set temperature when a heater is "paused"
    bool newTargetReached;				///< flag set when heater reached target and cleared when a new temperature is set

    uint16_t eeprom_base;               ///< Base address to read EEPROM configuration from

    PID pid;                            ///< PID controller instance
    bool bypassing_PID;                 ///< True if the heater is in full on

    bool fail_state;                    ///< True if the heater has detected a hardware
                                        ///< failure and is shut down.
    uint8_t fail_count;                 ///< Count of the number of hardware failures that
                                        ///< have been reported by #getTemperature().
                                        ///< If this goes over #SENSOR_MAX_BAD_READINGS,
                                        ///< then the heater will go into a fail state.
    uint8_t value_fail_count;			///< a second failure counter for valid temp reads that are out of range (eg too hot)
    HeaterFailMode fail_mode;			///< queryable state to indicate WHY the heater fails

    Timeout heatingUpTimer;				///< timeout indicating how long heater has been heating
    Timeout heatProgressTimer;			///< timeout to flag if heater is not heating up from start
    bool progressChecked;				///< flag that heating up progress has been checked.
    const bool heat_timing_check;       ///< allow disabling of heat progress timing for heated build platform.
    bool is_paused;						///< set to true when we wish to pause the heater from heating up
    bool is_disabled;					///< heaters are disabled when they are not present (user settable)

    // While the calibration offset is silly, we leverage the calibration_eeprom_offset
    // as a means of telling us if we're dealing with an extruder or HBP
    uint8_t calibration_eeprom_offset; //axis offset in HEATER_CALIBRATE
    //int8_t  calibration_offset;   // temperature offset for this heater in degrees C

    /// This is the interval between PID calculations.  It doesn't make sense for
    /// this to be fast (<1 sec) because of the long system delay between heater
    /// and sensor.
    const static micros_t UPDATE_INTERVAL_MICROS = 500L * 1000L;

    /// Put the heater into a failure state, ensuring that the heating element is
    /// disabled.
    void fail();

  public:
    /// Instantiate a new heater object.
    /// \param[in] sensor #TemperatureSensor element to use as an input
    /// \param[in] element #HeatingElement to use as an output
    /// \param[in] eeprom_base EEPROM address where the PID settings are stored.
    /// \param[in] heat_timing_check whether or not we should monitor heat-up time
    /// \param[in] calibration_offset axis offset in HEATER_CALIBRATE field of eeprom
    Heater(TemperatureSensor& sensor,
           HeatingElement& element,
           const uint16_t eeprom_base,
           bool heat_timing_check,
	   uint8_t calibration_offset);

    /// Get the current sensor temperature
    /// \return Current sensor temperature, in degrees Celcius
    int16_t get_current_temperature() { return current_temperature; }

    /// Get the setpoint temperature
    /// \return Setpoint temperature, in degrees Celcius
    int16_t get_set_temperature() { return ( is_paused ) ? paused_set_temperature : pid.getTarget(); }

    /// Set the target output temperature
    /// \param temp New target temperature, in degrees Celcius.
    void set_target_temperature(int temp);

    /// Check if the heater is within the specified band
    /// \return True if the heater temperature is within #TARGET_HYSTERESIS degrees
    ///         of the setpoint temperature.
    bool has_reached_target_temperature();

    /// Check if the heater is in a failure state
    /// \return true if the heater has failed.
    bool has_failed() { return fail_state; }

    /// Run the heater management loop. This must be called periodically
    void manage_temperature();

    /// Change the setpoint temperature
    /// \param value New setpoint temperature, in degrees Celcius.
    void set_output(uint8_t value);

    /// Reset the heater to a to board-on state
    void reset();

    /// clear heater target temp and states
    void abort();

    // pause or unpause the heater
    // a paused heater will not heat
    void Pause(bool on);

    bool isPaused() { return is_paused;}

#if defined(SUPPORT_GET_PID_STATE)
    /// Get the current PID error term
    /// \return E term from the PID controller
    int16_t getPIDErrorTerm() { return pid.getErrorTerm(); }

    /// Get the current PID delta term
    /// \return D term from the PID controller
    int16_t getPIDDeltaTerm() { return pid.getDeltaTerm(); }

    /// Get the last PID output
    /// \return last output from the PID controller
    int16_t getPIDLastOutput() { return pid.getLastOutput(); }
#endif

    /// get the difference between the current temperature
    /// and the set temperature
    int16_t getDelta();

    /// is heater temperature target different than current
    bool isHeating();

    /// is heater temperature target less than current temperature
    bool isCooling();

    /// get heater fail mode
    uint8_t GetFailMode() { return fail_mode; }

    void disable(bool on);

    bool isDisabled(){return is_disabled;}

#if defined(HAS_VIKI_INTERFACE)
     void viki_led(bool state);
#endif
};

#endif // HEATER_H
